package org.bdware.sc.crdt.planning;

public class PlanningWith1Expansivity extends SharableNetworkPlanning {
    public PlanningWith1Expansivity(String[] nodeIds, int[] writers, int[] readers, long maxDelay, int bandwidthDownload, int bandwidthUpload, int dataSize) {
        this.nodeIds = nodeIds;
        this.writers = writers;
        this.readers = readers;
        this.maxDelay = maxDelay;
        this.bandwidthDownload = bandwidthDownload;
        this.bandwidthUpload = bandwidthUpload;
        this.dataSize = dataSize;
        this.totalCountW = writers.length;
        this.totalCountR = readers.length;
    }

    public boolean writerTreeConstraint() {
        double common = frequencySyncW * (treeNodeCountW - 1) * dataSize;
        // 非叶子节点下载带宽
        boolean result1 = bandwidthDownload >= common;
        // 非根节点上行带宽
        boolean result2 = bandwidthUpload >= common / treeDegreeW;
        return result1 && result2;
    }

    public boolean writer2ReaderConstraint() {
        double common = frequencySyncWR * totalCountW * dataSize;
        // Writer根节点上行带宽
        boolean result1 = bandwidthUpload >= common * rootCountR / rootCountW;
        // Reader根节点下载带宽
        boolean result2 = bandwidthDownload >= common;
        return result1 && result2;
    }

    public boolean readerTreeConstraint() {
        double common = frequencySyncR * totalCountW * dataSize;
        // Reader非叶子节点上行带宽
        boolean result1 = bandwidthUpload >= common * treeDegreeR;
        // Reader非根节点下载带宽
        boolean result2 = bandwidthDownload >= common;
        return result1 && result2;
    }

    public void calcOptimizedResult() {
        double a = treeHeightW * Math.pow(treeDegreeW, treeHeightW) / (treeDegreeW - 1) + (treeDegreeW - Math.pow(treeDegreeW, treeHeightW + 1)) / Math.pow(treeDegreeW - 1, 2);
        double b = rootCountR * totalCountW;
        double c = (totalCountR - rootCountR) * totalCountW;

        double A = Math.sqrt(a * (treeHeightW - 1));
        double B = Math.sqrt(b);
        double C = Math.sqrt(c * (treeHeightR - 1));
        wDelay = (long) (maxDelay * (A / (A + B + C)));
        w2rDelay = (long) (maxDelay * (B / (A + B + C)));
        rDelay = (long) (maxDelay * (C / (A + B + C)));


        frequencySyncW = (treeHeightW - 1) / wDelay;
        frequencySyncR = (treeHeightR - 1) / rDelay;
        frequencySyncWR = 1.0 / w2rDelay;

        totalData = (long) (maxDelay * dataSize * (frequencySyncW * a + frequencySyncWR * b + frequencySyncR * c));
    }

    public static void main(String[] args) {
        long start = System.currentTimeMillis();
        String[] nodeIds = new String[1500];
        int[] writers = new int[1000];
        int[] readers = new int[1000];
        long maxDelay = 60;
        int bandwidthUpload = 10 * 1024 * 1024;
        int bandwidthDownload = 10 * 1024 * 1024;
        int datasize = 10 * 1024;

        PlanningWith1Expansivity planning = new PlanningWith1Expansivity(nodeIds, writers, readers, maxDelay, bandwidthDownload, bandwidthUpload, datasize);

        long minTotalData = Long.MAX_VALUE;
        String result = "";
        for (int rootCountW = 1; rootCountW <= writers.length; ++rootCountW) {
            for (int treeDegreeW = 2; treeDegreeW <= writers.length / rootCountW - 1; ++treeDegreeW) {
                planning.adjustWriterTree(rootCountW, treeDegreeW);
                for (int rootCountR = 1; rootCountR <= readers.length; ++rootCountR) {
                    for (int treeDegreeR = 2; treeDegreeR <= readers.length / rootCountR - 1; ++treeDegreeR) {
                        planning.adjustReaderTree(rootCountR, treeDegreeR);
                        planning.calcOptimizedResult();
                        if (!planning.readerTreeConstraint()) {
                            //System.out.println("reader");
                            continue;
                        }
                        if (!planning.writerTreeConstraint()) {
                            //System.out.println("writer");
                            continue;
                        }
                        if (!planning.writer2ReaderConstraint()) {
                            //System.out.println("writer2Reader");
                            continue;
                        }
                        if (minTotalData > planning.totalData) {
                            minTotalData = planning.totalData;
                            result = planning.toString();
                        }
                    }
                }
            }
        }

        System.out.println(minTotalData);
        System.out.println(result);
        long end = System.currentTimeMillis();
        System.out.println("took " + (end - start));
    }
}
