package org.bdware.sc.crdt.planning;

public class PlanningWithkExpansivity extends SharableNetworkPlanning {
    private double k;

    private double domainSize;

    public PlanningWithkExpansivity(String[] nodeIds, int[] writers, int[] readers, long maxDelay, int bandwidthDownload, int bandwidthUpload, int dataSize, double domainSize) {
        this.nodeIds = nodeIds;
        this.writers = writers;
        this.readers = readers;
        this.maxDelay = maxDelay;
        this.bandwidthDownload = bandwidthDownload;
        this.bandwidthUpload = bandwidthUpload;
        this.dataSize = dataSize;
        this.totalCountW = writers.length;
        this.totalCountR = readers.length;
        this.domainSize = domainSize;
        this.k = (domainSize - dataSize) / domainSize;
    }

    private double kWithPow(double v) {
        return Math.pow(k, (v - 1) / (treeDegreeW - 1));
    }

    public boolean writerTreeConstraint() {
        double common = frequencySyncW * domainSize * (1 - kWithPow(Math.pow(treeDegreeW, treeHeightW - 2)));
        // 非叶子节点下载带宽
        boolean result1 = bandwidthDownload >= common * treeDegreeW;
        // 非根节点上行带宽
        boolean result2 = bandwidthUpload >= common;
        return result1 && result2;
    }

    public boolean writer2ReaderConstraint() {
        double common = frequencySyncWR * domainSize * (1 - Math.pow(k, treeNodeCountW));
        // Writer根节点上行带宽
        boolean result1 = bandwidthUpload >= common * rootCountR;
        // Reader根节点下载带宽
        boolean result2 = bandwidthDownload >= common * rootCountW;
        return result1 && result2;
    }

    public boolean readerTreeConstraint() {
        double common = frequencySyncR * domainSize * (1 - Math.pow(k, totalCountW));
        // Reader非叶子节点上行带宽
        boolean result1 = bandwidthUpload >= common * treeDegreeR;
        // Reader非根节点下载带宽
        boolean result2 = bandwidthDownload >= common;
        return result1 && result2;
    }

    double[] accumulations = new double[1000];

    double calcAccumulation(int H1) {
        if (accumulations[H1] > 0) {
            return accumulations[H1];
        }
        double result = 0;
        for (int h = 1; h < H1; ++h) {
            result += (Math.pow(treeDegreeW, h) * (1 - kWithPow(Math.pow(treeDegreeW, H1 - h - 1))));
        }
        return result;
    }

    public void calcOptimizedResult() {
        double a = calcAccumulation((int) treeHeightW);
        double b = rootCountR * rootCountW * (1 - Math.pow(k, treeNodeCountW));
        double c = (totalCountR - rootCountR) * (1 - Math.pow(k, totalCountW));

        double A = Math.sqrt(a * (treeHeightW - 1));
        double B = Math.sqrt(b);
        double C = Math.sqrt(c * (treeHeightR - 1));
        wDelay = (long) (maxDelay * (A / (A + B + C)));
        w2rDelay = (long) (maxDelay * (B / (A + B + C)));
        rDelay = (long) (maxDelay * (C / (A + B + C)));


        frequencySyncW = (treeHeightW - 1) / wDelay;
        frequencySyncR = (treeHeightR - 1) / rDelay;
        frequencySyncWR = 1.0 / w2rDelay;

        totalData = (long) (maxDelay * domainSize * (frequencySyncW * a + frequencySyncWR * b + frequencySyncR * c));
    }

    public static void main(String[] args) {
        long start = System.currentTimeMillis();
        String[] nodeIds = new String[1500];
        int[] writers = new int[1000];
        int[] readers = new int[1000];
        long maxDelay = 60;
        int bandwidthUpload = 10 * 1024 * 1024;
        int bandwidthDownload = 10 * 1024 * 1024;
        int datasize = 10 * 1024;
        double domainSize = 1000 * 1024;

        PlanningWithkExpansivity planning = new PlanningWithkExpansivity(nodeIds, writers, readers, maxDelay, bandwidthDownload, bandwidthUpload, datasize, domainSize);

        planning.adjustAndCalc();

        long end = System.currentTimeMillis();
        System.out.println("took " + (end - start));
    }
}
